/* Tree */
/* $Id: tree.c,v 1.3 2003/11/08 19:17:13 bsp Exp $ */
/* ... */

#include <stdlib.h>
#include <string.h>
#include "ptxlist.h"

ptxtree *tree_error_ptr;

ptxtree *init_tree (void *data, size_t size)
{
	ptxtree *t;
	char *d;
	
	t = calloc (1, sizeof (ptxtree));
	if (!t)
	{
		tree_error_ptr = NULL;
		return NULL;
	}
	
	d = calloc (size, sizeof (char));
	if (!d)
	{
		free (t);
		tree_error_ptr = NULL;
		return NULL;
	}
	
	memcpy (d, data, size);
	
	t->data = (void *) d;
	return t;
}

void goto_root (ptxtree **t)
{
	if (!t || !*t) return;
	while ((*t)->up) *t = (*t)->up;
	while ((*t)->prev) *t = (*t)->prev;
}

void goto_parent (ptxtree **t)
{
	if (!t || !*t) return;
	if ((*t)->up) *t = (*t)->up;
}

void goto_child (ptxtree **t)
{
	if (!t || !*t) return;
	if ((*t)->down) *t = (*t)->down;
}

void goto_prev_sibling (ptxtree **t)
{
	if (!t || !*t) return;
	if ((*t)->prev) *t = (*t)->prev;
}

void goto_next_sibling (ptxtree **t)
{
	if (!t || !*t) return;
	if ((*t)->next) *t = (*t)->next;
}

void goto_first_sibling (ptxtree **t)
{
	if (!t || !*t) return;
	while ((*t)->prev) *t = (*t)->prev;
}

void goto_last_sibling (ptxtree **t)
{
	if (!t || !*t) return;
	while ((*t)->next) *t = (*t)->next;
}

void goto_first_child (ptxtree **t)
{
	if (!t || !*t || !(*t)->down) return;
	*t = (*t)->down;
	while ((*t)->prev) *t = (*t)->prev;
}

void goto_last_child (ptxtree **t)
{
	if (!t || !*t || !(*t)->down) return;
	*t = (*t)->down;
	while ((*t)->next) *t = (*t)->next;
}

int is_root (ptxtree *t)
{
	if (!t) return -1;
	if (t->up) return 0;
	else return 1;
}

int is_first_sibling (ptxtree *t)
{
	if (!t) return -1;
	if (t->prev) return 0;
	else return 1;
}

int is_last_sibling (ptxtree *t)
{
	if (!t) return -1;
	if (t->next) return 0;
	else return 1;
}

ptxtree *add_sibling (ptxtree *t, void *data, size_t size)
{
	ptxtree *tmp;
	char *d;
	
	tmp = calloc (1, sizeof (ptxtree));
	if (!tmp)
	{
		tree_error_ptr = t;
		return NULL;
	}
	
	d = calloc (size, sizeof (char));
	if (!d)
	{
		free (tmp);
		tree_error_ptr = t;
		return NULL;
	}
	
	memcpy (d, data, size);
	
	tmp->data = d;
	
	if (!t) return tmp;
	
	tmp->up = t->up;
	tmp->next = t->next;
	tmp->prev = t;
	t->next = tmp;
	
	return tmp;
}

ptxtree *add_child (ptxtree *t, void *data, size_t size)
{
	ptxtree *tmp;
	char *d;
	
	tmp = calloc (1, sizeof (ptxtree));
	if (!tmp)
	{
		tree_error_ptr = t;
		return NULL;
	}
	
	d = calloc (size, sizeof (char));
	if (!d)
	{
		free (tmp);
		tree_error_ptr = t;
		return NULL;
	}
	
	memcpy (d, data, size);
	
	tmp->data = d;
	
	if (!t) return tmp;
	
	tmp->up = t;
	
	if (!t->down) 
	{
		t->down = tmp;
		return tmp;
	}
	
	t = t->down;
	
	tmp->next = t->next;
	tmp->prev = t;
	
	if (t->next) t->next->prev = tmp;
	
	goto_first_sibling (&t);
	t->up->down = t;
	
	return tmp;
}

ptxtree *del_tree_element (ptxtree *t)
{
	ptxtree *tmp = NULL, *tmp1;
	
	if (t->down)
	{
		tmp = t->down;
		goto_first_sibling (&tmp);
		tmp->prev = t->prev;
		for (tmp->up = t->up; tmp; tmp = tmp->next) 
		    tmp->up = t->up;
		
		tmp->next = t->next;
		tmp = t->down;
		
		free (t->data);
		free (t);
		
		return tmp;
	}
	
	if (t->next) 
	{
		tmp = t->next;
		tmp->prev = t->prev;
	}
	
	if (t->prev)
	{
		tmp = t->prev;
		tmp->next = t->next;
	}
	
	if (t->up)
	{
		if (t->next) tmp1 = t->next;
		else tmp1 = t->prev;
		goto_first_sibling (&tmp1);
		t->up->down = tmp1;
		if (!tmp) tmp = t->up;
	}
	
	free (t->data);
	free (t);
	
	return tmp;
}

void del_tree (ptxtree *t)
{
	while (t) t = del_tree_element (t);
}

void *get_tree_element_data (ptxtree *t)
{
	if (!t) return NULL;
	return t->data;
}
